-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local im = ui_imgui

local C = {}

C.name = "Get Scenario Data (Mod - Gull)"
C.description = "Gets info from the given scenario file."
C.category = "once_instant"
C.color = im.ImVec4(1, 0.75, 0.1, 0.8)
C.pinSchema = {
  {dir = "in", type = "flow", name = "flow"},
  {dir = "in", type = "flow", name = "reset", impulse = "true"},
  {dir = "in", type = "string", name = "scenarioFile", description = "Scenario file path to load."},

  {dir = "out", type = "flow", name = "flow"},
  {dir = "out", type = "string", name = "title", description = "Scenario title."},
  {dir = "out", type = "string", name = "text", description = "Scenario description."},
  {dir = "out", type = "string", name = "author", hidden = true, description = "Scenario author."},
  {dir = "out", type = "number", name = "difficulty", hidden = true, description = "Scenario difficulty."}
}

function C:reset()
  self.pinOut.flow.value = false
end

function C:init()
  self:reset()
end

function C:_executionStopped()
  self:reset()
end

function C:work()
  if self.pinIn.reset.value then
    self:reset()
  end
  if not self.pinOut.flow.value and self.pinIn.scenarioFile.value then
    local data = jsonReadFile(self.pinIn.scenarioFile.value)
    if data then
      data = data[1] or data
      self.pinOut.title.value = data.name
      self.pinOut.text.value = data.description
      self.pinOut.author.value = data.authors
      self.pinOut.difficulty.value = data.difficulty
      self.pinOut.flow.value = true
    end
  end
end

return _flowgraph_createNode(C)