-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local im = ui_imgui

local C = {}

C.name = "Setup Race (Mod - Gull)"
C.description = "Loads the extension for the custom Race game mode."
C.category = "once_f_duration"
C.color = im.ImVec4(1, 0.75, 0.1, 0.8)
C.pinSchema = {
  {dir = "in", type = "flow", name = "flow"},
  {dir = "in", type = "flow", name = "reset", impulse = "true"},
  {dir = "in", type = "string", name = "extension", description = "Extension to load (format = folder_file)."},
  {dir = "in", type = "string", name = "seriesFile", description = "Race series.json file."},
  {dir = "in", type = "number", name = "raceIdx", description = "Race number to load."},

  {dir = "out", type = "flow", name = "flow"},
  {dir = "out", type = "flow", name = "ready", description = "Flows when the scene is ready for gameplay."},
  {dir = "out", type = "string", name = "scenarioFile", description = "Linked scenario file path."}
}

function C:reset()
  if self.extStr then
    extensions.unload(self.extStr)
  end
  self.extStr = nil
  self.pinOut.flow.value = false
  self.pinOut.ready.value = false
  self.pinOut.scenarioFile.value = nil
end

function C:init()
  self:reset()
end

function C:_executionStopped()
  self:reset()
end

function C:work()
  if self.pinIn.reset.value then
    self:reset()
  end
  if not self.pinOut.flow.value then
    if self.pinIn.extension.value and self.pinIn.seriesFile.value and self.pinIn.raceIdx.value then
      self.extStr = self.pinIn.extension.value
      if extensions[self.extStr] then
        extensions.reload(self.extStr)
      else
        extensions.load(self.extStr)
      end

      if extensions[self.extStr] then
        extensions[self.extStr].setupScenario(self.pinIn.seriesFile.value, self.pinIn.raceIdx.value)
        self.pinOut.scenarioFile.value = extensions[self.extStr].getScenarioFile()
      end
    end
    self.pinOut.flow.value = true
  else
    if self.extStr then
      local state = extensions[self.extStr].getState()
      if state == "ready" then
        self.pinOut.ready.value = true
      end
    end
  end
end

return _flowgraph_createNode(C)