local M = {}

local logTag = "raceExtras"

local player, racers, racePath, stats, finishers = {}, {}, {}, {}, {}
local mapNodes, aiMod, aiExt, finishPos, finishVec, finishRadius, achievements, currPrefabName
local optionsExt, raceUtils, raceMarkers, scenarioFile
local firstLoad = true
local validAi = true
local isNight = false
local state = "none"
local mainTimer, tickTimer, startTimer, finishTimer = 0, 0, 0, 0
local tickTime = 0.25
local aggroScale = 1
local racePathLength = 0
local vecUp = vec3(0, 0, 1)
local carLimit = Engine.Platform.getCPUInfo().coresPhysical or 4
local userCarLimit = core_settings_settings.getValue("trafficAmount")
local spawnProcess = {state = 0, delay = 0}
local ranks = {"1st", "2nd", "3rd", "4th", "5th", "6th", "7th", "8th", "9th", "10th", "11th", "12th"}
local seriesFile = "settings/raceSeries/series.json"
local series = {
  name = "default",
  tag = "race",
  saveFile = "settings/cloud/raceSeriesSave.json",
  achievementsFile = "",
  races = {},
  playerName = "Player",
  racerNames = {},
  messages = {
    win = "You won the race!",
    finish = "won the race.",
    place = "You placed",
    finishTime = "Time:",
    finishReward = "Reward:",
    money = "Total cash:",
    achievements = "Achievements:",
    congrats = "Perfect!",
    stuck = "You got stuck!",
    flip = "You flipped over!"
  },
  styles = {text = {}}
}
local extras = {
  useInfoTool = true,
  aiMod = "ai",
  damageCutoff = 10000,
  stuckTime = 3,
  crawlSpeed = 2,
  aggroValue = 1,
  aggroVariance = 0,
  aggroAlterRate = 0.02,
  aggroRubberband = 0,
  aggroRubberbandBounds = {-0.4, 0.05},
  avoidCars = false,
  rewards = {500, 250, 125, 60, 40, 20}
}

local function racesTable() -- races data
  local data = {}
  for i, v in ipairs(series.races) do
    data[v] = {wins = 0, finishes = 0, score = 0, time = 0}
    local scenarioData = jsonReadFile(extras.levelPath.."scenarios/"..v..".json")
    if scenarioData then
      data[v].name = scenarioData[1].name
      if series.tag then
        data[v].name = string.match(data[v].name, string.upper(series.tag).."[%s%p]*(.*)")
      end
    end
    if not data[v].name then data[v].name = "race"..i end
  end

  return data
end

local function achievementsTable() -- achievements data
  local data = {
    notDone = {},
    done = {},
    total = 0
  }
  if achievements then
    for k, v in pairs(achievements) do
      table.insert(data.notDone, k)
      data.total = data.total + 1
    end
  end

  return data
end

local function settingsTable() -- settings data
  local data = {
    difficulty = "Normal",
    disableFail = false,
    disableAchievements = false
  }
  return data
end

local function validateAchievements(saveAchData) -- validates the saved achievements table with the current series achievements
  if achievements then
    local achKeys = tableKeys(achievements)
    local saveAchDict = tableValuesAsLookupDict(arrayConcat(shallowcopy(saveAchData.done), shallowcopy(saveAchData.notDone)))
    saveAchData.total = #achKeys

    for i, v in ipairs(achKeys) do
      if not saveAchDict[v] then -- series achievement entry not found in save data
        table.insert(saveAchData.notDone, v)
      end
    end
    for i = #saveAchData.done, 1, -1 do
      local v = saveAchData.done[i]
      if not achievements[v] then -- save data achievement done entry not found in series achievements
        table.remove(saveAchData.done, i)
      end
    end
    for i = #saveAchData.notDone, 1, -1 do
      local v = saveAchData.notDone[i]
      if not achievements[v] then -- save data achievement not done entry not found in series achievements
        table.remove(saveAchData.notDone, i)
      end
    end
  end

  return saveAchData
end

local function setupSaveFile() -- creates and returns save data
  local json = {
    level = extras.levelPath,
    series = series.name
  }
  json.money = 0
  json.wins = 0
  json.finishes = 0
  json.score = 0
  json.minScore = 0
  json.time = 0
  json.distance = 0
  json.races = racesTable()
  json.achievements = achievementsTable()
  json.settings = settingsTable()
  return json
end

local function modifySaveFile(params) -- set specific save data values
  local json = jsonReadFile(series.saveFile)
  if params.resetStats then
    for k, v in pairs(json) do
      if type(v) == "number" then
        json[k] = 0
      end
    end
    json.races = racesTable()
  end
  if params.resetAchievements then
    json.achievements = achievementsTable()
  end
  if params.resetSettings then
    json.settings = settingsTable()
  end
  if params.settings then
    json.settings = tableMerge(json.settings, params.settings)
  end

  jsonWriteFile(series.saveFile, json)
end

local function getState() -- returns the main state
  return state
end

local function editScenarioValue(key, value) -- sets extras values manually
  extras[key] = value
end

local function getScenarioFile() -- returns the scenario file path
  return scenarioFile
end

local function createRacePath() -- creates the entire race path, as well as the modified AI path
  mapNodes = map.getMap().nodes
  local mapGp = map.getGraphpath()
  local n1, n2
  extras.lapCountAi = extras.lapCount > 1 and extras.lapCount + 1 or 1
  extras.waypointsAi = shallowcopy(extras.waypoints)
  local num = #extras.waypoints
  n1 = mapNodes[extras.waypoints[num - 1]] and extras.waypoints[num - 1] or map.findClosestRoad(vec3(scenetree.findObject(extras.waypoints[num - 1]):getPosition()))
  n2 = mapNodes[extras.waypoints[num]] and extras.waypoints[num] or map.findClosestRoad(vec3(scenetree.findObject(extras.waypoints[num]):getPosition()))
  local shortPath = mapGp:getPath(n1, n2, 1)
  num = #shortPath
  finishVec = (mapNodes[shortPath[num]].pos - mapNodes[shortPath[num - 1]].pos):z0():normalized()

  if extras.lapCount == 1 then
    local extendPath = mapGp:getRandomPathG(n2, finishVec, 300)
    if not extendPath[1] then
      log("W", logTag, "No extended path after finish found!")
    else
      table.insert(extras.waypointsAi, extendPath[#extendPath])
    end
  else
    table.insert(extras.waypointsAi, extras.waypoints[1])
  end
  n1, n2 = map.findClosestRoad(player.pos)
  if (mapNodes[n1].pos - player.pos):normalized():dot(player.dirVec) > 0 then
    n1, n2 = n2, n1
  end

  racePath = {{n1, 0}} -- {node, distance}
  for _, v in ipairs(extras.waypoints) do
    n2 = mapNodes[v] and v or map.findClosestRoad(vec3(scenetree.findObject(v):getPosition()))
    local shortPath = mapGp:getPath(n1, n2, 1)
    if not shortPath[1] then
      log("E", logTag, "Race path error! No path from "..n1.." to "..n2)
    end
    for i, w in ipairs(shortPath) do
      if i > 1 then
        racePathLength = racePathLength + mapNodes[shortPath[i - 1]].pos:distance(mapNodes[w].pos)
        table.insert(racePath, {w, racePathLength})
      end
    end
    n1 = n2
  end
  racePathLength = racePathLength * extras.lapCount
end

local function setupPlayer() -- sets up player
  local obj = be:getPlayerVehicle(0)
  player.id = obj:getID()
  player.name = obj.name
  player.pos = vec3(obj:getPosition())
  player.dirVec = vec3(obj:getDirectionVector())
  player.dirVecUp = vec3(obj:getDirectionVectorUp())
  player.role = "player"
  player.speed = 0
  player.damage = 0
  player.driveDist = 0
  player.width = obj.initialNodePosBB:getExtents().x
  player.length = obj.initialNodePosBB:getExtents().y
  player.path = {node = "", idx = 1, percent = 0}

  obj.playerUsable = true
  if series.playerName then
    core_vehicles.setPlateText(series.playerName, player.id)
    player.driver = series.playerName
  end
end

local function setupRacers(vehicles) -- sets up racers
  table.sort(vehicles, function(a, b) return a > b end)
  for i, id in ipairs(vehicles) do
    local r = {}
    local obj = be:getObjectByID(id)
    r.id = id
    r.name = obj.name
    r.pos = vec3(obj:getPosition())
    r.dirVec = vec3(obj:getDirectionVector())
    r.dirVecUp = vec3(obj:getDirectionVectorUp())
    r.role = "ai"
    r.speed = 0
    r.damage = 0
    r.path = {node = "", idx = 1, percent = 0}

    obj.playerUsable = false
    if aiMod ~= "ai" then
      obj:queueLuaCommand("extensions.load('"..aiMod.."')")
    end
    if extras.namesList[i] then
      core_vehicles.setPlateText(extras.namesList[i], r.id)
      r.driver = extras.namesList[i]
    end
    table.insert(racers, r)
  end
end

local function resetPlayer() -- resets player data
  local obj = be:getObjectByID(player.id)
  obj:queueLuaCommand("recovery.loadHome()")
  obj:queueLuaCommand("controller.setFreeze(1)")

  player.driveDist = player.driveDist + racePath[player.path.idx][2]
  player.path = {node = "", idx = 1, percent = 0}
  player.raceData = {currIdx = 0, currLap = 1, totalLaps = extras.lapCount, wheelOffsets = {}, currCorners = {}}

  -- wheel offsets and corners are used to have precise waypoint triggering
  local wheelCount = obj:getWheelCount()
  local vRot = quatFromDir(player.dirVec, player.dirVecUp)
  local x, y, z = vRot * vec3(1, 0, 0), vRot * vec3(0, 1, 0), vRot * vec3(0, 0, 1)

  for i = 1, wheelCount do
    local axisNodes = obj:getWheelAxisNodes(i - 1)
    local nodePos = vec3(obj:getNodePosition(axisNodes[1]))
    local pos = vec3(nodePos:dot(x), nodePos:dot(y), nodePos:dot(z))
    player.raceData.wheelOffsets[i] = pos
    player.raceData.currCorners[i] = vRot * pos + player.pos
  end
end

local function resetRacers() -- resets racers data
  for i, r in ipairs(racers) do
    r.hitFlag = false
    r.path = {node = "", idx = 1, percent = 0}

    local obj = be:getObjectByID(r.id)
    if not validAi then
      obj:queueLuaCommand("extensions.unload('"..aiMod.."')")
    end
    if aiExt ~= "ai." then
      obj:queueLuaCommand(aiExt.."reset()")
    end
    obj.playerUsable = false
    obj:queueLuaCommand("recovery.loadHome()")
    obj:queueLuaCommand(aiExt.."setMode('stop')")
    obj:queueLuaCommand(aiExt.."setAggression(1.4)")
    obj:queueLuaCommand("controller.setFreeze(1)")
  end
end

local function prepareRacers() -- sets up AI just before the start of the race
  for _, r in ipairs(racers) do
    local obj = be:getObjectByID(r.id)
    r.aggroOffset = extras.aggroVariance * (math.random() * 2 - 1)
    r.aggroBase = extras.aggroValue * aggroScale + r.aggroOffset
    r.aggro = r.aggroBase
    r.lapsDone = 0
    r.aggroAlt = math.max(1, r.aggro * 1.4)
    r.distTarget = 0

    if extras.forceLateralValue then
      r.forceLateral = extras.forceLateralVariance and extras.forceLateralValue + (math.random() * 2 - 1) * extras.forceLateralVariance or extras.forceLateralValue
      obj:queueLuaCommand(aiExt.."setForceLateral("..tostring(r.forceLateral)..")")
    end
    if extras.distPastEdgeValue then
      r.distPastEdge = extras.distPastEdgeVariance and extras.distPastEdgeValue + (math.random() * 2 - 1) * extras.distPastEdgeVariance or extras.distPastEdgeValue
      obj:queueLuaCommand(aiExt.."setDistPastEdge("..tostring(r.distPastEdge)..")")
    end
    obj:queueLuaCommand("controller.getControllerSafe('esc').setESCMode(2)")
    obj:queueLuaCommand(aiExt.."driveUsingPath({wpTargetList = "..serialize(extras.waypointsAi)..", noOfLaps = "..tostring(extras.lapCountAi)..", aggression = "..tostring(r.aggroAlt)..", avoidCars = "..tostring(extras.avoidCars).."})")
  end
end

local function resetStats() -- resets the stats table
  stats = {topSpeed = 0, damage = 0, maxRoll = 1, headStart = 0, leadTime = 0, maxLeadTime = 0, blockTime = 0, maxBlockTime = 0, finishTime = 0, finishPlace = 0, finishGap = math.huge, finishSpeed = math.huge, finishDrive = 1, halfwayPlace = 0, reboundDifference = 0, stuckRacers = {}, rolledRacers = {}, hitRacers = {}, racersCount = #racers}
end

local function adjustAggro(r) -- dynamically adjusts the aggression value of an AI racer, to help improve its ability
  if r.path.percent >= 10 and extras.aggroRubberband > 0 then
    local rb = extras.aggroRubberbandBounds
    r.aggro = clamp(r.aggroBase + clamp((player.path.percent - r.path.percent) / (20 / extras.aggroRubberband), rb[1], rb[2]), 0.3, 1.3)
  end

  if extras.watchNodes then -- preset node & aggression value pairs
    for i, v in ipairs(extras.watchNodes) do
      if racePath[r.path.idx + 1] and v[1] == racePath[r.path.idx + 1][1] then
        r.aggroAlt = clamp(v[2] * aggroScale + r.aggroOffset, 0.3, 1.3)
        be:getObjectByID(r.id):queueLuaCommand(aiExt.."setAggression("..tostring(r.aggroAlt)..")")
        break
      end
    end
  end

  if r.aggroAlt ~= r.aggro then -- aggression only gets set if value changes
    if r.aggroAlt > r.aggro then
      r.aggroAlt = math.max(r.aggro, r.aggroAlt - extras.aggroAlterRate)
    elseif r.aggroAlt < r.aggro then
      r.aggroAlt = math.min(r.aggro, r.aggroAlt + extras.aggroAlterRate)
    end
    be:getObjectByID(r.id):queueLuaCommand(aiExt.."setAggression("..tostring(r.aggroAlt)..")")
  end
end

local function trackPlace() -- tracks the player standings throughout the race
  local place = 1
  for _, r in ipairs(racers) do
    if r.path.percent > player.path.percent then
      place = place + 1
    end
  end
  player.place = place
end

local function trackProgress(r) -- tracks the race path progress of the given vehicle
  local bestIdx
  local bestDist = math.huge
  local nodeCount = #racePath

  for i = math.max(1, r.path.idx - 5), math.min(nodeCount - 1, r.path.idx + 5) do
    local n1, n2 = mapNodes[racePath[i][1]], mapNodes[racePath[i + 1][1]]
    if n1 and n2 then
      local roadDist = r.pos:squaredDistanceToLineSegment(n1.pos, n2.pos)
      if roadDist < bestDist then
        bestIdx = i
        bestDist = roadDist
      end
    end
  end

  if bestIdx then
    local p1, p2 = mapNodes[racePath[bestIdx][1]].pos, mapNodes[racePath[bestIdx + 1][1]].pos
    local offset = linePointFromXnorm(p1, p2, clamp(p1:distance(r.pos) / p1:distance(p2), 0, 1))
    offset = offset:distance(p1)
    local percent = ((racePath[bestIdx][2] + offset + racePath[nodeCount][2] * r.lapsDone) / racePathLength) * 100
    r.path = {node = racePath[bestIdx][1], idx = bestIdx, percent = percent}
  end
end

local function trackStats(tick) -- tracks race stats during the race
  local blockState = false
  stats.damage = player.damage
  if player.speed > stats.topSpeed and player.speed <= 100 then
    stats.topSpeed = player.speed
  end
  if stats.topSpeed <= extras.crawlSpeed then
    stats.headStart = mainTimer
  end
  if player.dirVecUpDot < stats.maxRoll then
    stats.maxRoll = player.dirVecUpDot
  end
  if player.place == 1 then
    stats.leadTime = stats.leadTime + tick
    if stats.leadTime > stats.maxLeadTime then
      stats.maxLeadTime = stats.leadTime
    end
  else
    stats.leadTime = 0
  end
  if stats.halfwayPlace == 0 and player.path.percent >= 50 then
    stats.halfwayPlace = player.place
  end

  for _, r in ipairs(racers) do
    if r.state == "race" then
      if r.dist <= player.length + 1 and r.path.percent < player.path.percent and (player.pos - r.pos):normalized():dot(player.dirVec) >= 0.95 then
        blockState = true
      end
      if r.dirVecUp:dot(vecUp) <= 0 and r.dist <= 100 then
        if player.dirVec:dot((r.pos - player.pos):normalized()) > 0 then
          if not tableContains(stats.rolledRacers, r.id) then
            table.insert(stats.rolledRacers, r.id)
          end
        end
      end
    end
  end

  if blockState then
    stats.blockTime = stats.blockTime + tick
    if stats.blockTime > stats.maxBlockTime then
      stats.maxBlockTime = stats.blockTime
    end
  else
    stats.blockTime = 0
  end
end

local function setFinishStats() -- tracks race stats at the instant the player passes the finish line
  local bestDist = math.huge

  for _, r in ipairs(racers) do
    if finishers[1] == "Player" then
      if r.dist < bestDist then
        bestDist = r.dist
      end
    end

    if r.speed <= extras.crawlSpeed then
      table.insert(stats.stuckRacers, r.id)
    end
    if r.hitFlag then
      table.insert(stats.hitRacers, r.id)
    end
  end

  stats.finishTime = mainTimer
  stats.reboundDifference = stats.halfwayPlace - stats.finishPlace
  stats.finishGap = bestDist
  stats.finishSpeed = player.speed
  stats.finishDrive = player.dirVec:dot(finishVec)

  stats.hitRacersCount = #stats.hitRacers
  stats.stuckRacersCount = #stats.stuckRacers
  stats.rolledRacersCount = #stats.rolledRacers
end

local function setPostFinishStats(data) -- tracks race stats after the race
  stats.money = data.money
  stats.wins = data.wins
  stats.distance = player.driveDist
  stats.totalScore = 0
  stats.minScore = 10
  stats.seriesTime = 0
  stats.seriesDistance = data.distance + player.driveDist

  for k, v in pairs(data.races) do
    stats.totalScore = stats.totalScore + v.score
    if v.score < stats.minScore then
      stats.minScore = v.score
    end

    if stats.seriesTime < 1e6 and v.score > 0 then
      stats.seriesTime = stats.seriesTime + v.time
    else
      stats.seriesTime = 1e6
    end
  end
end

local function checkAchievements(saveAchData) -- checks and saves achievements
  local playerWin = finishers[1] == "Player"

  if achievements then
    for k, v in pairs(achievements) do
      if not v.needsWin or (v.needsWin and playerWin) then
        if not tableContains(saveAchData.done, k) and raceUtils.checkCondition(v.cond, stats) then
          local notDoneIdx = arrayFindValueIndex(saveAchData.notDone, k)
          if notDoneIdx then
            table.remove(saveAchData.notDone, notDoneIdx)
          end
          table.insert(saveAchData.done, k)

          local data = {}
          data.key = k
          data.image = "achievement"
          data.time = 8
          data.subtitle = v.title
          data.text = {v.desc}

          optionsExt.insertQueue(data)
        end
      end
    end
  end

  return saveAchData
end

local function insertNewPaths(data) -- inserts custom road segments into the map and reloads the map
  local mapFile = extras.levelPath.."map.json"
  local json = jsonReadFile(mapFile)
  data.segments = data.segments or {}

  if json then
    if data.fileName then
      local json = jsonReadFile(data.fileName)
      if json then
        data.segments = tableMerge(data.segments, json.segments)
      end
    end
    json.segments = tableMerge(json.segments, data.segments)

    jsonWriteFile(extras.levelPath.."map.json", json, true)
    log("I", logTag, "Inserted custom data into map.json")
  end
end

local function spawnPlayer(playerData) -- spawns the player vehicle
  log("I", logTag, "Spawning player vehicle...")

  local options = {}
  local colorKeys = {"color1", "color2", "color3"}
  local info = playerData.data
  local model = info.model
  local config = "vehicles/"..model.."/"..info.config..".pc"

  for i, key in ipairs({"paint", "paint2", "paint3"}) do
    if info[colorKeys[i]] then
      local c = stringToTable(info[colorKeys[i]])
      options[key] = createVehiclePaint({x = c[1], y = c[2], z = c[3], w = c[4]})
    end
  end

  playerData.pos = vec3(playerData.pos)
  playerData.rot = quat(playerData.rot)

  spawn.spawnVehicle(model, config, playerData.pos, playerData.rot, options)
end

local function spawnRacers(racersData) -- spawns the racers vehicle group
  log("I", logTag, "Spawning racer vehicles...")

  if racersData.minQuantity then
    carLimit = carLimit - 1
    if userCarLimit and userCarLimit > 0 then
      carLimit = userCarLimit
    end
    racersData.quantity = math.min(11, clamp(carLimit, racersData.minQuantity, racersData.quantity))
  end

  core_multiSpawn.setGroup("race", racersData.data)
  core_multiSpawn.setupVehicles(racersData.quantity, racersData.order, racersData.mode, racersData.gap)
end

local function setupInfoTool() -- sets up the imgui race tool
  local data = jsonReadFile(series.saveFile)
  if not data then
    log("E", logTag, "Unable to read save data!")
    return
  end

  data.width = 560
  data.height = 480
  data.image = "info"
  data.tag = series.tag
  data.levelPath = extras.levelPath
  data.racesList = series.races
  data.difficultyList = {"Easy", "Normal"}
  data.paints = core_vehicles.getModel(be:getPlayerVehicle(0).jbeam).model.paints
  data.paintKeys = tableKeys(data.paints or {})
  data.extraFeatures = (state == "ready" and data.wins >= 3)

  if series.styles and series.styles.text then
    data.title = series.styles.text.title
    data.subtitle = series.styles.text.subtitle
  end
  table.sort(data.paintKeys)
  data.achievementsDoneList = shallowcopy(data.achievements.done)
  data.achievementsNotDoneList = shallowcopy(data.achievements.notDone)
  data.achievementsRef = achievements
  table.sort(data.achievementsDoneList)
  table.sort(data.achievementsNotDoneList)

  optionsExt.activate(data, "mini")
end

local function setupScenario(seriesFile, idx) -- main init function, sets up the core of the scenario
  state = "none"
  extensions.load("core_multiSpawn")

  player, racers, racePath, stats, finishers = {}, {}, {}, {}, {}
  core_multiSpawn.deleteVehicles()
  if be:getPlayerVehicle(0) then
    be:getPlayerVehicle(0):delete()
  end

  seriesFile = seriesFile or ""
  idx = idx or 0
  series = jsonReadFile(seriesFile) or series

  local optionsStr = "gameplay_"..series.tag.."_options"
  local utilsStr = "/extensions/gameplay/"..series.tag.."/raceUtils"
  extensions.load("gameplay_"..series.tag.."_options")
  raceUtils = require(utilsStr)
  optionsExt = extensions[optionsStr]

  achievements = jsonReadFile(series.achievementsFile)

  local filePath = path.split(seriesFile)
  local scenarioKey = series.tag.."_race"..idx
  scenarioFile = filePath..scenarioKey..".json"
  local json = jsonReadFile(scenarioFile)
  if not json then
    log("E", logTag, "Unable to read scenario data!")
    return
  end
  json = json[1]

  if currPrefabName then raceUtils.unloadPrefab(currPrefabName) end
  raceUtils.loadPrefab(scenarioKey, filePath..scenarioKey..".prefab")
  currPrefabName = scenarioKey

  local wpCount = 0
  for _, name in ipairs(scenetree.findClassObjects("BeamNGWaypoint")) do
    if string.find(name, json.waypointPrefix or "noPrefix") then
      wpCount = wpCount + 1
    end
  end

  if wpCount == 0 then
    log("E", logTag, "No matching waypoints found!")
    return
  end

  extras.waypoints = {}
  for i = 1, wpCount do
    table.insert(extras.waypoints, json.waypointPrefix..i)
  end

  extras.key = scenarioKey
  extras.name = json.name
  extras.levelPath = path.split(getMissionFilename(), "(.-)([^/]-([^%.]*))$")
  extras.lapCount = json.lapCount or 1
  extras.lapPoint = extras.waypoints[#extras.waypoints]
  extras.namesList = arrayShuffle(series.racerNames)

  if json.timeOfDay then
    core_environment.setTimeOfDay({time = json.timeOfDay})
  end
  if json.manualPaths then
    insertNewPaths(json.manualPaths)
  end
  if json.spawn then
    spawnProcess.player = json.spawn.player
    spawnProcess.racers = json.spawn.racers
    spawnProcess.state = 1
  end
  if json.extras then
    extras = tableMerge(extras, json.extras)
  end

  extras.watchNodes = json.watchNodes
  if series.styles then
    optionsExt.setStyles(series.styles)
  end
  extras.messages = series.messages or extras.messages

  be:reloadCollision()
  map.reset()

  state = "loading"
end

local function clearScenario() -- cleans up the scenario
  bullettime.set(1)
  if raceUtils then
    raceUtils.unloadPrefab(currPrefabName or "")
  end
  if optionsExt then
    optionsExt.activate()
  end
  if extras.levelPath then
    os.remove(extras.levelPath.."map.json")
  end
end

local function onVehicleResetted(id)
  if isNight then
    local command = id == player.id and "electrics.setLightsState(2)" or "electrics.setLightsState(1)" -- low beams for AI vehicles
    be:getObjectByID(id):queueLuaCommand(command)
  end
end

local function onResetGameplay()
  if player.id then
    be:getObjectByID(player.id):resetBrokenFlexMesh()
    be:getObjectByID(player.id):reset()
  end
  guihooks.trigger("ChangeState", "menu")
end

local function onRaceReady()
  state = "ready"

  mainTimer, tickTimer, startTimer, finishTimer = 0, 0, 0, 0
  aiMod = extras.aiMod or "ai"
  aiExt = validAi and aiMod.."." or "ai."

  isNight = false
  local day = core_environment.getTimeOfDay()
  if day and day.time and day.time >= 0.23 and day.time <= 0.77 then
    isNight = true
  end

  resetPlayer()
  resetRacers()
  resetStats()

  local newMarkers, raceData = raceUtils.setupMarkers(extras.waypoints, extras.lapCount)
  raceMarkers = newMarkers
  player.raceData = tableMerge(player.raceData, raceData)
  bullettime.set(1)

  if firstLoad then
    local json = jsonReadFile(series.saveFile)
    if not json or json.series ~= series.name then
      json = setupSaveFile()
    end
    if not json.settings then
      json.settings = settingsTable()
    end

    json.achievements = validateAchievements(json.achievements)
    extras = tableMerge(extras, json.settings)
    jsonWriteFile(series.saveFile, json)

    firstLoad = false
  end

  if extras.useInfoTool then
    setupInfoTool()
  end
end

local function onCountdownStarted()
  state = "countdown"
  player.timers = {stuck = 0, flip = 0}
  player.lapsDone = 0
  player.place = 0
  finishers = {}
  commands.setGameCamera()

  if extras.useInfoTool then
    optionsExt.activate(nil, "hidden")
  end
end

local function onRaceStart()
  state = "race"

  raceMarkers, player.raceData = raceUtils.updateMarkers(raceMarkers, player.raceData)
  guihooks.trigger("WayPoint", "Checkpoint 0 / "..tostring(#extras.waypoints))
  if extras.lapCount > 1 then
    guihooks.trigger("RaceLapChange", {current = 1, count = extras.lapCount})
  else
    guihooks.trigger("RaceLapChange", nil)
  end
  be:getObjectByID(player.id):queueLuaCommand("controller.setFreeze(0)")

  for _, r in ipairs(racers) do
    r.state = "race"
    be:getObjectByID(r.id):queueLuaCommand("controller.setFreeze(0)")
  end
end

local function onRaceTick(tick)
  local obj = be:getObjectByID(player.id)
  player.dirVecUpDot = player.dirVecUp:dot(vecUp)
  player.speed = vec3(obj:getVelocity()):length()
  player.damage = map.objects[player.id].damage or 0

  if not extras.disableFail then
    if player.timers.flip >= extras.stuckTime then
      raceUtils.finishScreen(extras.name, extras.messages.flip, true, mainTimer)
      state = "finish"
    end
    if player.timers.stuck >= extras.stuckTime then
      raceUtils.finishScreen(extras.name, extras.messages.stuck, true, mainTimer)
      state = "finish"
    end
  end
  if player.speed <= extras.crawlSpeed then
    if player.dirVecUpDot <= 0.1 then
      player.timers.flip = player.timers.flip + tick
    else
      player.timers.flip = 0
    end
    if player.timers.flip == 0 and player.damage >= extras.damageCutoff then
      player.timers.stuck = player.timers.stuck + tick
    end
  else
    player.timers.stuck = 0
  end

  local avg = 0
  for _, r in ipairs(racers) do
    obj = be:getObjectByID(r.id)
    r.dirVec:set(obj:getDirectionVector())
    r.dirVecUp:set(obj:getDirectionVectorUp())
    r.speed = vec3(obj:getVelocity()):length()
    r.damage = map.objects[r.id].damage
    r.dist = r.pos:distance(player.pos)
    trackProgress(r)
    if r.state == "race" then
      adjustAggro(r)
    end
    avg = avg + r.speed
  end
  avg = avg / math.max(1, #racers)
  if math.ceil(mainTimer) == 2 and avg <= 1 then -- AI is not moving, try to use default AI module
    validAi = false
  end

  trackProgress(player)
  trackPlace()
  trackStats(tick)
end

local function onRaceResult()
  state = "finish"

  local finishText = ""
  local finishPlace = 0
  local errorText = "Error: Invalid race stats!"
  local reward, medalScore, raceScore = 0, 0, 0
  for i, v in ipairs(finishers) do
    if v == "Player" then
      finishPlace = i
      medalScore = math.max(0, 4 - i)
      raceScore = math.max(1, 11 - i)
    end
  end

  if finishers[1] then
    if finishers[1] == "Player" then
      finishText = extras.messages.win
    else
      finishText = finishers[1].." "..extras.messages.finish
      finishText = finishText.." "..extras.messages.place.." "..ranks[finishPlace].."."
    end
    if extras.rewards[finishPlace] then
      local multiplier = aggroScale == 1 and 1 or 0.5
      reward = math.ceil(extras.rewards[finishPlace] * multiplier)
    end
    local json = jsonReadFile(series.saveFile)
    local race = json.races[extras.key]
    json.money = json.money + reward
    json.finishes = json.finishes + 1
    race.score = math.max(raceScore, race.score)
    race.finishes = race.finishes + 1
    if race.time == 0 or mainTimer <= race.time then
      race.time = math.floor(mainTimer * 100) / 100
    end
    if raceScore == 10 then
      json.wins = json.wins + 1
      race.wins = race.wins + 1
    end

    stats.finishPlace = finishPlace
    setPostFinishStats(json)

    json.time = math.floor(stats.seriesTime * 100) / 100
    json.distance = math.floor(stats.seriesDistance)
    json.score = stats.totalScore
    json.minScore = stats.minScore

    finishText = finishText.." "..extras.messages.finishTime.." "..raceUtils.timeToString(mainTimer).."."
    finishText = finishText.." "..extras.messages.finishReward..reward.."."
    finishText = finishText.." "..extras.messages.money..json.money.."."

    if not extras.disableAchievements then
      json.achievements = checkAchievements(json.achievements)
      stats.achievementsDone = #json.achievements.done
      stats.achievementsTotal = stats.achievementsDone + #json.achievements.notDone

      finishText = finishText.." "..extras.messages.achievements.." "..stats.achievementsDone.."/"..stats.achievementsTotal.." ."
      if stats.achievementsDone >= stats.achievementsTotal then
        finishText = finishText.." "..extras.messages.congrats
      end
    end
    if json.money >= 1000000 or json.money < 0 then
      finishText = finishText.." It looks like you cheated with your money. Yoink, it's all gone now!"
      json.money = 0
    end
    if extras.playerSetVehicle then
      json.playerSetVehicle = true
    end
    player.driveDist = 0

    jsonWriteFile(series.saveFile, json)
  end

  if stats.finishPlace == 0 then
    raceUtils.finishScreen(extras.name, errorText, true, mainTimer)
    log("W", logTag, "Invalid race finish state!")
  else
    local screenStats = {score = {label = "score", points = medalScore, maxPoints = 3, relativePoints = medalScore / 3 * 100}}
    if stats.finishPlace <= 3 then
      raceUtils.finishScreen(extras.name, finishText, false, mainTimer, screenStats)
    else
      raceUtils.finishScreen(extras.name, finishText, true, mainTimer, screenStats)
    end
  end
end

local function onRaceWaypointReached(data)
  if data.vehicleId == player.id then
    raceMarkers, player.raceData = raceUtils.updateMarkers(raceMarkers, player.raceData)
    guihooks.trigger("WayPoint", "Checkpoint "..tostring(player.raceData.currIdx - 1).." / "..tostring(#extras.waypoints))
    Engine.Audio.playOnce("AudioGui", "event:UI_Checkpoint")

    if data.waypointName == extras.lapPoint then
      player.lapsDone = player.lapsDone + 1
      player.path.idx = 1

      if player.lapsDone >= extras.lapCount then -- player finish
        table.insert(finishers, "Player")
        state = "transition"
        player.driveDist = player.driveDist + racePathLength
        guihooks.trigger("ScenarioFlashMessage", {{ranks[#finishers], 2, 0, true}})
        setFinishStats()
      else
        if extras.lapCount > 1 then
          guihooks.trigger("RaceLapChange", {current = player.raceData.currLap, count = extras.lapCount})
        end
      end
    end
  end
end

local function onVehicleSpawned(id)
  if state == "loading" then
    if be:getObjectCount() == 1 then
      be:enterVehicle(0, be:getObjectByID(id))
      commands.setGameCamera()
    end
    spawnProcess.delay = 3
  end

  if player.id and state == "ready" then
    if id ~= player.id then
      be:getObjectByID(id).group = "other"
      be:enterVehicle(0, be:getObjectByID(player.id))
    end
    be:getObjectByID(player.id):queueLuaCommand("controller.setFreeze(1)")
    extras.playerSetVehicle = true
    extensions.hook("onResetGameplay")
  end
end

local function onVehicleGroupSpawned(vehIds)
  for _, v in ipairs(vehIds) do
    be:getObjectByID(v):queueLuaCommand("recovery.saveHome()") -- saves post-adjustment positions
  end

  aiMod = extras.aiMod or "ai"
  aiExt = validAi and aiMod.."." or "ai."
  setupPlayer()
  setupRacers(vehIds)
  finishPos = vec3(scenetree.findObject(extras.waypoints[#extras.waypoints]):getPosition())
  finishRadius = getSceneWaypointRadius(scenetree.findObject(extras.waypoints[#extras.waypoints]))
  createRacePath()
  resetRacers()
end

local function setAggroScale(value) -- sets the base aggro scale (pseudo difficulty)
  if not value then value = 1 end
  aggroScale = value
end

local function doDebug() -- debug visuals
  if racePath then
    local nodeCount = #racePath
    for i, node in ipairs(racePath) do
      local c = rainbowColor(nodeCount, i, 1)
      local r = i > 1 and 0.6 or 1.2
      c[4] = 0.6
      debugDrawer:drawSphere((mapNodes[node[1]].pos + vec3(0, 0, i / 1000)):toPoint3F(), r, ColorF(unpack(c)))
    end
  end

  for _, r in ipairs(racers) do
    debugDrawer:drawTextAdvanced((r.pos + vecUp):toPoint3F(), String(tostring(math.floor(r.aggroAlt * 100) / 100)), ColorF(1, 1, 1, 1), true, false, ColorI(0, 0, 0, 255))
  end
  if player.id then
    local vel = vec3(be:getObjectByID(player.id):getVelocity())
    if state == "race" and player.damage >= 40000 then
      table.insert(finishers, "Player")
      onRaceResult()
    end
  end
end

local function onUpdate(dt, dtSim)
  if state ~= "none" and not editor.active then
    if spawnProcess.state > 0 and spawnProcess.state < 3 then
      if spawnProcess.state == 1 and spawnProcess.player then
        spawnPlayer(spawnProcess.player)
        spawnProcess.player = nil
      end
      if spawnProcess.state == 2 and spawnProcess.racers then
        spawnRacers(spawnProcess.racers)
        spawnProcess.racers = nil
      end

      spawnProcess.delay = spawnProcess.delay - 1
      if spawnProcess.delay == 0 then
        spawnProcess.state = spawnProcess.state + 1
        if spawnProcess.state == 3 then
          state = "ready"
        end
      end
    end

    if player.id then
      local pObj = be:getObjectByID(player.id)
      player.pos:set(pObj:getPosition())
      player.dirVec:set(pObj:getDirectionVector())
      player.dirVecUp:set(pObj:getDirectionVectorUp())

      if player.raceData then
        for i, corner in ipairs(player.raceData.wheelOffsets) do
          local vRot = quatFromDir(player.dirVec, player.dirVecUp)
          player.raceData.currCorners[i] = player.pos + vRot * corner
        end

        local wpObj = scenetree.findObject(extras.waypoints[player.raceData.currIdx] or "nil")
        if wpObj then
          local pos = vec3(wpObj:getPosition())
          local radius = vec3(wpObj:getScale()).z

          for _, cPos in ipairs(player.raceData.currCorners) do
            if cPos:squaredDistance(pos) < square(radius) then
              extensions.hook("onRaceWaypointReached", {vehicleId = player.id, waypointName = extras.waypoints[player.raceData.currIdx]})
              break
            end
          end
        end
      end
    end

    if racers[1] then
      for _, r in ipairs(racers) do
        if aiExt ~= "ai." then
          be:getObjectByID(r.id):queueLuaCommand(aiExt.."updateGFX("..tostring(dt)..")")
        end
        r.pos:set(be:getObjectByID(r.id):getPosition())
        r.distTarget = r.pos:squaredDistance(finishPos)
        if r.state == "race" then
          if not r.hitFlag and map.objects[r.id] and map.objects[r.id].objectCollisions[player.id] == 1 then
            r.hitFlag = true
          end
          if r.distTarget <= square(finishRadius) then
            if r.path.idx >= 3 then
              r.lapsDone = r.lapsDone + 1
              r.path.idx = 1
            end
            if r.lapsDone >= extras.lapCount then -- racer finish
              be:getObjectByID(r.id):queueLuaCommand(aiExt.."setAggression(0.25)")
              local str = r.driver or "Racer #"..r.id
              table.insert(finishers, str)
              r.state = "finish"
            end
          end
        end
      end
    end

    if state == "countdown" then
      startTimer = startTimer + dtSim
      if startTimer >= 2 then
        prepareRacers()
        startTimer = -1e6
      end
    end

    if state == "race" then
      mainTimer = mainTimer + dtSim
      tickTimer = tickTimer + dtSim
      guihooks.trigger("raceTime", {time = mainTimer})

      if tickTimer >= tickTime then
        onRaceTick(tickTime)
        tickTimer = tickTimer - tickTime
      end
    elseif state == "transition" then
      finishTimer = finishTimer + dtSim
      if finishTimer >= 2.5 then
        for _, r in ipairs(racers) do
          be:getObjectByID(r.id).playerUsable = true
        end
        onRaceResult()
        finishTimer = -1e6
      end
    end
  end

  if M.debug then doDebug() end
end

local function onPreRender(dt, dtSim)
  if raceMarkers then
    raceMarkers.render(dt, dtSim)
  end
end

local function onRaceSeriesSettingsChanged(data)
  modifySaveFile(data)

  if data.paint then
    raceUtils.setVehPaint(player.id, data.paint)
  end
  if data.difficulty then
    local value = v == "Easy" and 0.75 or 1
    setAggroScale(value)
  end
  if data.settings then
    extras = tableMerge(extras, data.settings)
  end
end

local function onRaceToolRequest()
  bullettime.set(1)
  core_input_actionFilter.clear(0)
  guihooks.trigger("ChangeState", {state = "menu"})
  setupInfoTool()
end

local function onExtensionLoaded()
end

local function onExtensionUnloaded()
  clearScenario()
end

local function onVehicleDestroyed()
  clearScenario()
end

local function onExit()
  clearScenario()
end

M.setupScenario = setupScenario
M.clearScenario = clearScenario
M.setupInfoTool = setupInfoTool
M.getScenarioFile = getScenarioFile
M.setupSaveFile = setupSaveFile
M.modifySaveFile = modifySaveFile
M.getState = getState
M.edit = editScenarioValue

M.onRaceSeriesSettingsChanged = onRaceSeriesSettingsChanged
M.onRaceToolRequest = onRaceToolRequest

M.onVehicleResetted = onVehicleResetted
M.onResetGameplay = onResetGameplay
M.onRaceReady = onRaceReady
M.onCountdownStarted = onCountdownStarted
M.onRaceStart = onRaceStart
M.onRaceResult = onRaceResult
M.onRaceWaypointReached = onRaceWaypointReached
M.onVehicleSpawned = onVehicleSpawned
M.onVehicleGroupSpawned = onVehicleGroupSpawned
M.onUpdate = onUpdate
M.onPreRender = onPreRender
M.onExtensionLoaded = onExtensionLoaded
M.onExtensionUnloaded = onExtensionUnloaded
M.onVehicleDestroyed = onVehicleDestroyed
M.onExit = onExit

return M